/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

/*
 * IDE support
 */
#include <mpc8xx_irq.h>
#include <ppcboot.h>
#include <command.h>
#include <ide.h>
#include <ata.h>
#include <cmd_ide.h>
#include <mpc8xx.h>

#if (CONFIG_COMMANDS & CFG_CMD_IDE)

/* ------------------------------------------------------------------------- */

/*
 * Allow configuration to select PCMCIA slot,
 * or try to generate a useful default
 */
#if !defined(CONFIG_PCMCIA_SLOT_A) && !defined(CONFIG_PCMCIA_SLOT_B)

					/* The RPX series use SLOT_B	*/
#if defined(CONFIG_RPXCLASSIC) || defined(CONFIG_RPXLITE)
# define CONFIG_PCMCIA_SLOT_B
#elif defined(CONFIG_ADS)		/* The ADS  board use SLOT_A	*/
# define CONFIG_PCMCIA_SLOT_A
#elif defined(CONFIG_FADS)		/* The FADS series are a mess	*/
# if defined(CONFIG_MPC860T) || defined(CONFIG_MPC860) || defined(CONFIG_MPC821)
#  define CONFIG_PCMCIA_SLOT_A
# else
#  define CONFIG_PCMCIA_SLOT_B
# endif
#elif defined(CONFIG_TQM860L) || defined(CONFIG_TQM855L) /* The TQM8xxL modules */
# define	CONFIG_PCMCIA_SLOT_A	/* ... use SLOT_A on MPC860/855	*/
#elif defined(CONFIG_TQM823L) || defined(CONFIG_TQM850L)
# define	CONFIG_PCMCIA_SLOT_B	/* ... and SLOT_B else		*/
#elif defined(CONFIG_SPD823TS)		/* The SPD8xx use SLOT_B	*/
# define CONFIG_PCMCIA_SLOT_B
#else
# error "PCMCIA Slot not configured"
#endif

#endif /* !defined(CONFIG_PCMCIA_SLOT_A) && !defined(CONFIG_PCMCIA_SLOT_B) */

/* Make sure exactly one slot is defined - we support only one for now */
#if !defined(CONFIG_PCMCIA_SLOT_A) && !defined(CONFIG_PCMCIA_SLOT_B)
#error Neither CONFIG_PCMCIA_SLOT_A nor CONFIG_PCMCIA_SLOT_B configured
#endif
#if defined(CONFIG_PCMCIA_SLOT_A) && defined(CONFIG_PCMCIA_SLOT_B)
#error Both CONFIG_PCMCIA_SLOT_A and CONFIG_PCMCIA_SLOT_B configured
#endif

#define PCMCIA_SOCKETS_NO	1
#define PCMCIA_MEM_WIN_NO	4
#define PCMCIA_IO_WIN_NO	2

/* define _slot_ to be able to optimize macros */
#ifdef CONFIG_PCMCIA_SLOT_A
# define _slot_			0
# define PCMCIA_SLOT_MSG	"SLOT_A"
#else
# define _slot_			1
# define PCMCIA_SLOT_MSG	"SLOT_B"
#endif

/*
 * The TQM850L hardware has two pins swapped! Grrrrgh!
 */
#ifdef	CONFIG_TQM850L
#define __MY_PCMCIA_GCRX_CXRESET	PCMCIA_GCRX_CXOE
#define __MY_PCMCIA_GCRX_CXOE		PCMCIA_GCRX_CXRESET
#else
#define __MY_PCMCIA_GCRX_CXRESET	PCMCIA_GCRX_CXRESET
#define __MY_PCMCIA_GCRX_CXOE		PCMCIA_GCRX_CXOE
#endif

/* look up table for pgcrx registers */

static u_int *pcmcia_pgcrx[2] = {
	&((immap_t *)CFG_IMMR)->im_pcmcia.pcmc_pgcra,
	&((immap_t *)CFG_IMMR)->im_pcmcia.pcmc_pgcrb,
};

#define PCMCIA_PGCRX(slot)	(*pcmcia_pgcrx[slot])

/*
 * This structure is used to address each window in the PCMCIA controller.
 *
 * Keep in mind that we assume that pcmcia_win_t[n+1] is mapped directly
 * after pcmcia_win_t[n]...
 */

typedef struct {
	ulong	br;
	ulong	or;
} pcmcia_win_t;

/*
 * LED Port
 */
#define LED_IDE2	0x02
#define LED_IDE1	0x01

/* ------------------------------------------------------------------------- */

/*
 * Memory Map Information:
 *
 * BR0 / OR0:	EPROM1	Base 0x0C000000	Size 512KB
 * BR1 / OR1:	EPROM2	Base 0x0C080000	Size 512KB
 * BR2 / OR2:	SRAM	Base 0x04200000	Size   2MB
 * BR3 / OR3:	SDRAM	Base 0x00000000	Size  16MB
 * BR4 / OR4:	PER-8	Base 0x04000000	Size   1MB
 * BR5 / OR5:	SHARC	Base 0x04400000	Size   4MB
 */
/* ------------------------------------------------------------------------- */

/* Current I/O Device	*/
static int curr_device;

/* Current offset for IDE0 / IDE1 bus access	*/
static ulong bus_offset[IDE_MAXBUS] = {
	CFG_ATA_IDE0_OFFSET,
	CFG_ATA_IDE1_OFFSET,
};

#define	ATA_CURR_BASE(dev)	(ATA_IO_BASE+bus_offset[IDE_BUS(dev)])

typedef struct ide_dev_id {
	ulong	size;
	uchar	model[40];
	uchar	serial_no[20];
} ide_dev_id_t;

static ide_dev_id_t ide_device[IDE_MAXDEVICE];

/* ------------------------------------------------------------------------- */

static void  ide_led   (uchar led, uchar status);
static void  ide_reset (void);
static void  ide_ident (int device);
static void  ide_print (int device);
static uchar ide_wait  (int dev, ulong t);

static void __inline__ outb(int dev, int port, unsigned char val);
static unsigned char __inline__ inb(int dev, int port);
static void input_swap_data(int dev, ulong *sect_buf, int words);
static void input_data(int dev, ulong *sect_buf, int words);
static void output_data(int dev, ulong *sect_buf, int words);
static void trim_trail (unsigned char *str, unsigned int len);
/* ------------------------------------------------------------------------- */

void do_ide (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    switch (argc) {
    case 0:
    case 1:
	printf ("Usage:\n%s\n", cmdtp->usage);
	return;
    case 2:
	if (strcmp(argv[1],"reset") == 0) {
		printf ("\nReset IDE on PCMCIA " PCMCIA_SLOT_MSG ": ");

		ide_init();
		return;
	} else if (strcmp(argv[1],"info") == 0) {
		int i;

		printf ("\n");

		for (i=0; i<IDE_MAXDEVICE; ++i) {
			printf ("IDE device %d: ", i);
			ide_print (i);
		}
		return;

	} else if (strcmp(argv[1],"device") == 0) {
		printf ("\nIDE device %d: ", curr_device);
		ide_print (curr_device);
		return;
	}
	printf ("Usage:\n%s\n", cmdtp->usage);
	return;
    case 3:
	if (strcmp(argv[1],"device") == 0) {
		int dev = (int)simple_strtoul(argv[2], NULL, 10);

		printf ("\nIDE device %d: ", dev);
		if (dev >= IDE_MAXDEVICE) {
			printf ("unknown device\n");
			return;
		}

		ide_print (dev);

		if (ide_device[dev].size == 0) {
			return;
		}

		curr_device = dev;

		printf ("... is now current device\n");

		return;
	}
	printf ("Usage:\n%s\n", cmdtp->usage);
	return;
    default:
	/* at least 4 args */

	if (strcmp(argv[1],"read") == 0) {
		ulong addr = simple_strtoul(argv[2], NULL, 16);
		ulong blk  = simple_strtoul(argv[3], NULL, 16);
		ulong cnt  = simple_strtoul(argv[4], NULL, 16);
		ulong n;

		printf ("\nIDE read: device %d block # %ld, count %ld ... ",
			curr_device, blk, cnt);

		n = ide_read (0, blk, cnt, (ulong *)addr);

		printf ("%ld blocks read: %s\n",
			n, (n==cnt) ? "OK" : "ERROR");

		return;

	} else if (strcmp(argv[1],"write") == 0) {
		ulong addr = simple_strtoul(argv[2], NULL, 16);
		ulong blk  = simple_strtoul(argv[3], NULL, 16);
		ulong cnt  = simple_strtoul(argv[4], NULL, 16);
		ulong n;

		printf ("\nIDE write: device %d block # %ld, count %ld ... ",
			curr_device, blk, cnt);

		n = ide_write (0, blk, cnt, (ulong *)addr);

		printf ("%ld blocks written: %s\n",
			n, (n==cnt) ? "OK" : "ERROR");

		return;
	} else {
		printf ("Usage:\n%s\n", cmdtp->usage);
	}

	return;
    }
}

/* ------------------------------------------------------------------------- */

void ide_init (void)
{
	immap_t *immr = (immap_t *)CFG_IMMR;
	volatile pcmconf8xx_t *pcmp = &(immr->im_pcmcia);
	unsigned char c;
	int i;

	/* Configure PC for IDE Reset Pin
	 */
	immr->im_ioport.iop_pcdat &= ~(PC_IDE_RESET);	/* Set reset bit */
	immr->im_ioport.iop_pcpar &= ~(PC_IDE_RESET);
	immr->im_ioport.iop_pcso  &= ~(PC_IDE_RESET);
	immr->im_ioport.iop_pcdir |=   PC_IDE_RESET;	/* Make output	*/

	/* Reset the IDE just to be sure.
	 * Light LED's to show
	 */
	ide_led ((LED_IDE1 | LED_IDE2), 1);		/* LED's on	*/
	ide_reset ();
	ide_led ((LED_IDE1 | LED_IDE2), 0);		/* LED's off	*/

	/* PCMCIA / IDE initialization for common mem space */
	pcmp->pcmc_pgcrb = 0;

	/* IDE 1
	 */
	pcmp->pcmc_pbr0 = CFG_PCMCIA_PBR0;
	pcmp->pcmc_por0 = CFG_PCMCIA_POR0;

	pcmp->pcmc_pbr1 = CFG_PCMCIA_PBR1;
	pcmp->pcmc_por1 = CFG_PCMCIA_POR1;

	pcmp->pcmc_pbr2 = CFG_PCMCIA_PBR2;
	pcmp->pcmc_por2 = CFG_PCMCIA_POR2;

	pcmp->pcmc_pbr3 = CFG_PCMCIA_PBR3;
	pcmp->pcmc_por3 = CFG_PCMCIA_POR3;

	/* IDE 2
	 */
	pcmp->pcmc_pbr4 = CFG_PCMCIA_PBR4;
	pcmp->pcmc_por4 = CFG_PCMCIA_POR4;

	pcmp->pcmc_pbr5 = CFG_PCMCIA_PBR5;
	pcmp->pcmc_por5 = CFG_PCMCIA_POR5;

	pcmp->pcmc_pbr6 = CFG_PCMCIA_PBR6;
	pcmp->pcmc_por7 = CFG_PCMCIA_POR6;

	pcmp->pcmc_pbr7 = CFG_PCMCIA_PBR7;
	pcmp->pcmc_por7 = CFG_PCMCIA_POR7;

	/*
	 * Wait for IDE to get ready.
	 * According to spec, this can take up to 31 seconds!
	 */
	i = 0;
	do {
		udelay (10000);		/* 10 ms */
		c = inb (0, ATA_STATUS);
		++i;
		if (i > (ATA_RESET_TIME * 100)) {
			printf ("** Timeout **\n");
			return;
		}
		if ((i >= 100) && ((i%100)==0)) {
			printf (".");
		}
	} while (c & (ATA_STAT_BUSY | ATA_STAT_FAULT));

	if (c & (ATA_STAT_BUSY | ATA_STAT_FAULT)) {
		printf ("Status: 0x%02x\n", c);
	} else {
		printf ("OK\n");
	}

	curr_device = -1;
	for (i=0; i<IDE_MAXDEVICE; ++i) {
		ide_ident (i);
		ide_print (i);

		/* make first available device current */
		if ((ide_device[i].size > 0) && (curr_device < 0)) {
			curr_device = i;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void __inline__
outb(int dev, int port, unsigned char val)
{
	/* Ensure I/O operations complete */
	__asm__ volatile("eieio");
	*((uchar *)(ATA_CURR_BASE(dev)+port)) = val;
#if 0
printf ("\nOUTB: 0x%08lx <== 0x%02x\n", ATA_CURR_BASE(dev)+port, val);
#endif
}

static unsigned char __inline__
inb(int dev, int port)
{
	/* Ensure I/O operations complete */
	__asm__ volatile("eieio");
	return (*((uchar *)(ATA_CURR_BASE(dev)+port)));
}

static void
input_swap_data(int dev, ulong *sect_buf, int words)
{
	volatile ushort	*pbuf = (ushort *)(ATA_CURR_BASE(dev)+ATA_DATA_REG);
	ushort	*dbuf = (ushort *)sect_buf;

	while (words--) {
		*dbuf++ = ld_le16(pbuf);
		*dbuf++ = ld_le16(pbuf);
	}
}

static void
output_data(int dev, ulong *sect_buf, int words)
{
	ushort	*dbuf;
	volatile ushort	*pbuf;

	pbuf = (ushort *)(ATA_CURR_BASE(dev)+ATA_DATA_REG);
	dbuf = (ushort *)sect_buf;
	while (words--) {
		__asm__ volatile ("eieio");
		*pbuf = *dbuf++;
		__asm__ volatile ("eieio");
		*pbuf = *dbuf++;
	}
}

static void
input_data(int dev, ulong *sect_buf, int words)
{
	ushort	*dbuf;
	volatile ushort	*pbuf;

	pbuf = (ushort *)(ATA_CURR_BASE(dev)+ATA_DATA_REG);
	dbuf = (ushort *)sect_buf;
	while (words--) {
		__asm__ volatile ("eieio");
		*dbuf++ = *pbuf;
		__asm__ volatile ("eieio");
		*dbuf++ = *pbuf;
	}
}

/* ------------------------------------------------------------------------- */

static void ide_ident (int device)
{
	ulong iobuf[ATA_SECTORWORDS];
	unsigned char c;
	hd_driveid_t *iop = (hd_driveid_t *)iobuf;
	ide_dev_id_t *idp = &(ide_device[device]);

	printf ("    Device %d: ", device);

	/* Select device
	 */
	outb (device, ATA_DEV_HD, ATA_LBA | ATA_DEVICE(device));

	/* Start Ident Command
	 */
	outb (device, ATA_COMMAND, ATA_CMD_IDENT);

	/* Wait for completion
	 */
	c = ide_wait (device, 100);

	if ((c & ATA_STAT_READY) == 0) {
		idp->size = 0;
		idp->model[0] = idp->serial_no[0] = '\0';
		return;
	}

	input_swap_data (device, iobuf, ATA_SECTORWORDS);

	trim_trail (iop->model,     sizeof(iop->model));
	trim_trail (iop->serial_no, sizeof(iop->serial_no));

	/* swap shorts */
	idp->size = (iop->lba_capacity << 16) | (iop->lba_capacity >> 16);

	strncpy (idp->model,     iop->model,     sizeof(idp->model));
	strncpy (idp->serial_no, iop->serial_no, sizeof(idp->serial_no));
}

/* ------------------------------------------------------------------------- */

static void ide_print (int device)
{
	ide_dev_id_t *idp = &(ide_device[device]);

	if (idp->size == 0) {
		printf ("not available\n");
		return;
	}

	printf ("Model: %s  Serial #: %s  ", idp->model, idp->serial_no);
	printf ("Capacity: %lu MB = %lu GB\n",
		(idp->size / 2) >> 10, (idp->size / 2) >> 20);
}

/* ------------------------------------------------------------------------- */

ulong ide_read (int device, ulong blknr, ulong blkcnt, ulong *buffer)
{
	ulong n = 0;
	unsigned char c;

	while (blkcnt-- > 0) {

		c = ide_wait (device, 500);

		if (c & ATA_STAT_BUSY) {
			printf ("IDE read: device %d not ready\n", device);
			return (n);
		}

		outb (device, ATA_SECT_CNT, 1);
		outb (device, ATA_LBA_LOW,  (blknr >>  0) & 0xFF);
		outb (device, ATA_LBA_MID,  (blknr >>  8) & 0xFF);
		outb (device, ATA_LBA_HIGH, (blknr >> 16) & 0xFF);
		outb (device, ATA_DEV_HD,   ATA_LBA | ATA_DEVICE(device));
		outb (device, ATA_COMMAND,  ATA_CMD_READ);

		udelay (50);

		c = ide_wait (device, 500);	/* can't take over 500 ms */

		if ((c&(ATA_STAT_DRQ|ATA_STAT_BUSY|ATA_STAT_ERR)) != ATA_STAT_DRQ) {
			printf ("Error (no IRQ) dev %d blk %ld: status 0x%02x\n",
				device, blknr, c);
			return (n);
		}

		input_data (device, buffer, ATA_SECTORWORDS);
		(void) inb (device, ATA_STATUS);	/* clear IRQ */

		++n;
		++blknr;
		buffer += ATA_SECTORWORDS;
	}
	return (n);
}

/* ------------------------------------------------------------------------- */


ulong ide_write (int device, ulong blknr, ulong blkcnt, ulong *buffer)
{
	ulong n = 0;
	unsigned char c;

	while (blkcnt-- > 0) {

		c = ide_wait (device, 500);

		if (c & ATA_STAT_BUSY) {
			printf ("IDE read: device %d not ready\n", device);
			return (n);
		}

		outb (device, ATA_SECT_CNT, 1);
		outb (device, ATA_LBA_LOW,  (blknr >>  0) & 0xFF);
		outb (device, ATA_LBA_MID,  (blknr >>  8) & 0xFF);
		outb (device, ATA_LBA_HIGH, (blknr >> 16) & 0xFF);
		outb (device, ATA_DEV_HD,   ATA_LBA | ATA_DEVICE(device));
		outb (device, ATA_COMMAND,  ATA_CMD_WRITE);

		udelay (50);

		c = ide_wait (device, 500);	/* can't take over 500 ms */

		if ((c&(ATA_STAT_DRQ|ATA_STAT_BUSY|ATA_STAT_ERR)) != ATA_STAT_DRQ) {
			printf ("Error (no IRQ) dev %d blk %ld: status 0x%02x\n",
				device, blknr, c);
			return (n);
		}

		output_data (device, buffer, ATA_SECTORWORDS);
		c = inb (device, ATA_STATUS);	/* clear IRQ */
		++n;
		++blknr;
		buffer += ATA_SECTORWORDS;
	}
	return (n);
}

/* ------------------------------------------------------------------------- */

/* Trim trailing blanks, and NUL-terminate string
 */
static void trim_trail (unsigned char *str, unsigned int len)
{
	unsigned char *p = str + len - 1;

	while (len-- > 0) {
		*p-- = '\0';
		if (*p != ' ') {
			return;
		}
	}
}

/* ------------------------------------------------------------------------- */

/*
 * Wait until Busy bit is off, or timeout (in ms)
 * Return last status
 */
static uchar ide_wait (int dev, ulong t)
{
	ulong delay = 10 * t;		/* poll every 100 us */
	uchar c;

	while ((c = inb(dev, ATA_STATUS)) & ATA_STAT_BUSY) {
		udelay (100);
		if (delay-- == 0) {
			break;
		}
	}
	return (c);
}

/* ------------------------------------------------------------------------- */

static void ide_reset (void)
{
	immap_t *immr = (immap_t *)CFG_IMMR;

	/* assert IDE RESET signal */
	immr->im_ioport.iop_pcdat &= ~(PC_IDE_RESET);
	udelay (1000);
	/* de-assert RESET signal of IDE */
	immr->im_ioport.iop_pcdat |=   PC_IDE_RESET;
	udelay (1000);
}

/* ------------------------------------------------------------------------- */

static	uchar	led_buffer = 0;		/* Buffer for current LED status	*/

static void ide_led (uchar led, uchar status)
{
	uchar *led_port = (uchar *)(0x04000000 + 0x3000);

	if (status)	{		/* switch LED on	*/
		led_buffer |=  led;
	} else {			/* switch LED off	*/
		led_buffer &= ~led;
	}

	*led_port = led_buffer;
}

/* ------------------------------------------------------------------------- */

#endif	/* CONFIG_COMMANDS & CFG_CMD_IDE */
