/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef	_IDE_H
#define _IDE_H

/*
 * I/O Port definition for IDE Reset
 */
#define	PC_IDE_RESET		((ushort)0x0008)        /* PC 12 !!! */

/* Timings for IDE Interface - assuming 50 MHz clock !!!
 *
 * SETUP / LENGTH / HOLD - TIME AT 50 MHZ CLK
 * 70      165      30     PIO-Mode 0, [ns]
 *  4        9       2                 [Cycles]
 * 50      125      20     PIO-Mode 1, [ns]
 *  3        7       2                 [Cycles]
 * 30      100      15     PIO-Mode 2, [ns]
 *  2        6       1                 [Cycles]
 * 30       80      10     PIO-Mode 3, [ns]
 *  2        5       1                 [Cycles]
 * 25       70      10     PIO-Mode 4, [ns]
 *  2        4       1                 [Cycles]
 */

/*
 * Definitions for PCMCIA control registers to operate in IDE mode
 */

/* Window 0:
 *	Base: 0x04100000	CS1
 *	Port Size:     2 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:    16 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR0		0x04100000
#define CFG_PCMCIA_POR0	    (	PCMCIA_BSIZE_2	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_16	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 1:
 *	Base: 0x04100080	CS1
 *	Port Size:     8 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:     8 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR1		0x04100080
#define CFG_PCMCIA_POR1	    (	PCMCIA_BSIZE_8	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_8	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 2:
 *	Base: 0x04100100	CS2
 *	Port Size:     8 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:     8 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR2		0x04100100
#define CFG_PCMCIA_POR2	    (	PCMCIA_BSIZE_8	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_8	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 3:
 *	not used
 */
#define	CFG_PCMCIA_PBR3		0
#define	CFG_PCMCIA_POR3		0

/* Window 4:
 *	Base: 0x041000C00	CS1
 *	Port Size:     2 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:    16 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR4		0x04100C00
#define CFG_PCMCIA_POR4	    (	PCMCIA_BSIZE_2	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_16	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 5:
 *	Base: 0x04100C80	CS1
 *	Port Size:     8 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:     8 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR5		0x04100C80
#define CFG_PCMCIA_POR5	    (	PCMCIA_BSIZE_8	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_8	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 6:
 *	Base: 0x04100D00	CS2
 *	Port Size:     8 Bytes
 *	Strobe Hold:   2 clk	=> PIO Mode 0
 *	Strobe Setup:  4 clk	=> PIO Mode 0
 *	Strobe Length: 9 clk	=> PIO Mode 0
 *	Port Size:     8 Bit
 *	Common Memory Space
 */

#define	CFG_PCMCIA_PBR6		0x04100D00
#define CFG_PCMCIA_POR6	    (	PCMCIA_BSIZE_8	\
			    |	PCMCIA_SHT(2)	\
			    |	PCMCIA_SST(4)	\
			    |	PCMCIA_SL(9)	\
			    |	PCMCIA_PPS_8	\
			    |	PCMCIA_PRS_MEM	\
			    |	PCMCIA_PSLOT_B	\
			    |	PCMCIA_PV	\
			    )

/* Window 7:
 *	not used
 */
#define	CFG_PCMCIA_PBR7		0
#define	CFG_PCMCIA_POR7		0


#define	ATA_IO_BASE	CFG_PCMCIA_PBR0

#if 0	/* disabled - problems accessing IDE1 at 0x04100c00 */
#define	IDE_MAXBUS	2		/* max. 2 IDE busses		*/
#else
#define	IDE_MAXBUS	1		/* max. 2 IDE busses		*/
#endif
#define	IDE_MAXDEVICE	(IDE_MAXBUS*2)	/* max. 2 drives per IDE bus	*/
#define	IDE_BUS(dev)	(dev >> 1)

/*
 * Function Prototypes
 */

void  ide_init  (void);
ulong ide_read  (int device, ulong blknr, ulong blkcnt, ulong *buffer);
ulong ide_write (int device, ulong blknr, ulong blkcnt, ulong *buffer);

#endif /* _IDE_H */
